import sys
if sys.version_info[0] < 3:
    pass
#    import version
#    import util
#    import verror
#    import signaltype
#    import scope
#    import op2mark
#    import inference
#    import transition
